@@ -1,26 +1,36 @@
+import copy
+
 import torch
 from torch import nn
 import itertools
 from collections import defaultdict
 import numpy as np
 import math
-from .utils import slice_to_array
+from .utils import slice_to_index, logger, hash_tensor
+from .optim import BeamOptimizer
+from functools import partial
+import random
+import types
 
 
 class PackedSet(object):
 
-    def __init__(self, data, length=None):
+    def __init__(self, data, length=None, device=None):
 
         if length is None:
-            self.length = torch.LongTensor([0] + [len(x) for x in data])
             self.data = torch.cat(data, dim=0)
+            self.length = torch.LongTensor([0] + [len(x) for x in data], device=self.data.device)
         else:
-            self.length = torch.LongTensor([0] + list(length))
             self.data = data
+            self.length = torch.LongTensor([0] + list(length), device=self.data.device)
 
         self._offset = self.length.cumsum(dim=0)
         self.length = self.length[1:]
         self.offset = self._offset[:-1]
+        self.index = torch.arange(len(self.offset), device=self.data.device)
+
+        if device is not None:
+            self.to(device)
 
     def __len__(self):
         return len(self.offset)
@@ -33,6 +43,7 @@ class PackedSet(object):
         self._offset = self._offset.to(device)
         self.length = self.length.to(device)
         self.offset = self.offset.to(device)
+        self.index = self.index.to(device)
         return self
 
     def aggregate(self, func):
@@ -40,12 +51,17 @@ class PackedSet(object):
 
     def __getitem__(self, index):
 
-        index = slice_to_array(index)
-        if issubclass(type(index), np.ndarray):
-            index = torch.LongTensor(index)
+        index = slice_to_index(index, l=len(self))
+        if isinstance(index, np.ndarray):
+            if index.dtype == np.dtype('bool'):
+                index = torch.BoolTensor(index)
+            else:
+                index = torch.LongTensor(index)
         elif type(index) is int:
             index = torch.scalar_tensor(index, dtype=torch.int64)
-        if issubclass(type(index), torch.Tensor):
+        if isinstance(index, torch.Tensor):
+            if index.dtype == torch.bool:
+                index = self.index[index]
             shape = index.shape
             if len(shape) == 0:
                 return self.data[self._offset[index]:self._offset[index + 1]]
@@ -90,11 +106,12 @@ class PositionalHarmonicExpansion(object):
 
 class LinearNet(nn.Module):
 
-    def __init__(self, l_in, l_h=256, l_out=1, n_l=2, bias=True):
+    def __init__(self, l_in, l_h=256, l_out=1, n_l=2, bias=True, activation='ReLU'):
         super().__init__()
 
-        self.lin = nn.Sequential(*([nn.Linear(l_in, l_h, bias=bias), nn.ReLU()] if n_l > 1 else []),
-                                 *sum([[nn.Linear(l_h, l_h, bias=bias), nn.ReLU()] for _ in range(max(n_l - 2, 0))],
+        activation = getattr(nn, activation)
+        self.lin = nn.Sequential(*([nn.Linear(l_in, l_h, bias=bias), activation()] if n_l > 1 else []),
+                                 *sum([[nn.Linear(l_h, l_h, bias=bias), activation()] for _ in range(max(n_l - 2, 0))],
                                       []),
                                  nn.Linear(l_h if n_l > 1 else l_in, l_out, bias=bias))
 
@@ -104,137 +121,10 @@ class LinearNet(nn.Module):
         return y.squeeze(1)
 
 
-class MultipleScheduler(object):
-
-    def __init__(self, multiple_optimizer, scheduler, *argc, **argv):
-
-        self.schedulers = {}
-        self.multiple_optimizer = multiple_optimizer
-
-        for op in multiple_optimizer.optimizers.keys():
-            self.schedulers[op] = scheduler(multiple_optimizer.optimizers[op], *argc, **argv)
-
-    def step(self, *argc, **argv):
-        for op in self.multiple_optimizer.optimizers.keys():
-            self.schedulers[op].step(*argc, **argv)
-
-
-class BeamOptimizer(object):
-
-    def __init__(self, net, dense_args=None, clip=0, accumulate=1, amp=False,
-                 sparse_args=None, dense_optimizer='AdamW', sparse_optimizer='SparseAdam'):
-
-        sparse_optimizer = getattr(torch.optim, sparse_optimizer)
-        dense_optimizer = getattr(torch.optim, dense_optimizer)
-
-        if dense_args is None:
-            dense_args = {'lr': 1e-3, 'eps': 1e-4}
-        if sparse_args is None:
-            sparse_args = {'lr': 1e-2, 'eps': 1e-4}
-
-        self.clip = clip
-        self.accumulate = accumulate
-        self.iteration = 0
-        self.amp = amp
-        if self.amp:
-            self.scaler = torch.cuda.amp.GradScaler()
-
-        self.optimizers = {}
-
-        sparse_parameters = []
-        dense_parameters = []
-
-        for m in net.modules():
-            is_sparse = BeamOptimizer.check_sparse(m)
-            if is_sparse:
-                for p in m.parameters():
-                    if not any([p is pi for pi in sparse_parameters]):
-                        sparse_parameters.append(p)
-            else:
-                for p in m.parameters():
-                    if not any([p is pi for pi in dense_parameters]):
-                        dense_parameters.append(p)
-
-        if len(dense_parameters) > 0:
-            self.optimizers['dense'] = dense_optimizer(dense_parameters, **dense_args)
-
-        if len(sparse_parameters) > 0:
-            self.optimizers['sparse'] = sparse_optimizer(sparse_parameters, **sparse_args)
-
-        for k, o in self.optimizers.items():
-            setattr(self, k, o)
-
-    @staticmethod
-    def check_sparse(m):
-        return (issubclass(type(m), nn.Embedding) or issubclass(type(m), nn.EmbeddingBag)) and m.sparse
-
-    def set_scheduler(self, scheduler, *argc, **argv):
-        return MultipleScheduler(self, scheduler, *argc, **argv)
-
-    def reset(self):
-        self.iteration = 0
-        for op in self.optimizers.values():
-            op.state = defaultdict(dict)
-
-        self.zero_grad(set_to_none=True)
-        self.scaler = torch.cuda.amp.GradScaler() if self.amp else None
-
-    def zero_grad(self, set_to_none=True):
-        for op in self.optimizers.values():
-            op.zero_grad(set_to_none=set_to_none)
-
-    def apply(self, loss, training=False, set_to_none=True):
-
-        self.iteration += 1
-
-        if training:
-
-            if self.amp:
-                self.scaler.scale(loss).backward()
-            else:
-                loss.backward()
-
-            if self.clip > 0:
-                for op in self.optimizers.values():
-                    for pg in op.param_groups:
-
-                        if self.amp:
-                            for op in self.optimizers.values():
-                                self.scaler.unscale_(op)
-
-                        torch.nn.utils.clip_grad_norm_(iter(pg['params']), self.clip)
-
-            if not (self.iteration % self.accumulate):
-                self.step()
-                self.zero_grad(set_to_none=set_to_none)
-
-    def step(self):
-        for op in self.optimizers.values():
-            if self.amp:
-                self.scaler.step(op)
-            else:
-                op.step()
-
-        if self.amp:
-            self.scaler.update()
-
-    def state_dict(self):
-        return {k: op.state_dict() for k, op in self.optimizers.items()}
-
-    def load_state_dict(self, state_dict, state_only=False):
-
-        for k, op in self.optimizers.items():
-
-            if state_only:
-                state_dict[k]['param_groups'] = op.state_dict()['param_groups']
-
-            op.load_state_dict(state_dict[k])
-
-
 class RuleLayer(nn.Module):
 
     def __init__(self, n_rules, e_dim_in, e_dim_out, bias=True, pos_enc=None, dropout=0.0):
-        super(RuleLayer, self).__init__()
+        super().__init__()
 
         self.query = nn.Parameter(torch.empty((n_rules, e_dim_out)))
         nn.init.kaiming_uniform_(self.query, a=math.sqrt(5))
@@ -271,7 +161,7 @@ class MHRuleLayer(nn.Module):
 
     def __init__(self, n_rules, n_features, e_dim_out, bias=True, pos_enc=None, dropout=0.0, n_head=8,
                  static_attention=False):
-        super(MHRuleLayer, self).__init__()
+        super().__init__()
 
         self.query = nn.Parameter(torch.empty((n_rules, e_dim_out)))
         nn.init.kaiming_uniform_(self.query, a=math.sqrt(5))
@@ -293,6 +183,7 @@ class MHRuleLayer(nn.Module):
     def forward(self, x):
         b, nf, ne = x.shape
         pos = self.pos_enc.unsqueeze(0).repeat(b, 1, 1)
+
         x = x + pos
         v = self.value(x)
 
@@ -316,7 +207,7 @@ class MHRuleLayer(nn.Module):
 class LazyLinearRegression(nn.Module):
 
     def __init__(self, quantiles, lr=0.001):
-        super(LazyLinearRegression, self).__init__()
+        super().__init__()
 
         self.lr = lr
         self.quantiles = quantiles
@@ -347,10 +238,9 @@ class LazyLinearRegression(nn.Module):
 
 class SplineEmbedding(nn.Module):
 
-    def __init__(self, n_features, n_quantiles, emb_dim, n_tables=1, enable=True, init_weights=None):
-        super(SplineEmbedding, self).__init__()
+    def __init__(self, n_features, n_quantiles, emb_dim, n_tables=1, enable=True, init_weights=None, sparse=False):
+        super().__init__()
 
-        self.n_quantiles = n_quantiles
         self.n_features = n_features
         self.emb_dim = emb_dim
         self.enable = enable
@@ -358,9 +248,10 @@ class SplineEmbedding(nn.Module):
         self.n_tables = n_tables
         self.n_emb = (n_quantiles + 2) * n_features
         self.register_buffer('ind_offset', (n_quantiles + 2) * torch.arange(n_features, dtype=torch.int64).unsqueeze(0))
+        self.register_buffer("n_quantiles", torch.FloatTensor([n_quantiles]))
 
         if init_weights is None:
-            self.emb = nn.Embedding(self.n_emb * n_tables, emb_dim, sparse=False)
+            self.emb = nn.Embedding(self.n_emb * n_tables, emb_dim, sparse=sparse)
         else:
 
             none_val, base_1, base_0 = init_weights
@@ -370,17 +261,18 @@ class SplineEmbedding(nn.Module):
 
             weights = torch.cat([none_val.unsqueeze(2), weights], dim=2).reshape(-1, emb_dim)
 
-            self.emb = nn.Embedding.from_pretrained(weights, sparse=False, freeze=False)
+            self.emb = nn.Embedding.from_pretrained(weights, sparse=sparse, freeze=False)
 
     def forward(self, x, mask, rand_table):
 
-        if self.enable:
-            if not self.n_features:
-                return torch.FloatTensor(len(x), 0, self.emb_dim).to(x.device)
-        else:
-            return torch.zeros(*x.shape, self.emb_dim).to(x.device)
+        # if self.enable:
+        #     if not self.n_features:
+        #         return torch.Tensor(len(x), 0, self.emb_dim, device=x.device, dtype=x.dtype)
+        # else:
+        #     return torch.zeros(*x.shape, self.emb_dim, device=x.device, dtype=x.dtype)
 
-        x = torch.clamp(x, min=1e-6, max=1 - 1e-6)
+        th = 1e-3 if x.dtype == torch.float16 else 1e-6
+        x = torch.clamp(x, min=th, max=1-th)
 
         offset = self.ind_offset + self.n_emb * rand_table
 
@@ -402,7 +294,8 @@ class SplineEmbedding(nn.Module):
 class PID(object):
 
     def __init__(self, k_p=0.05, k_i=0.005, k_d=0.005, T=20, clip=0.005):
-        super(PID, self).__init__()
+        super().__init__()
+
         self.k_p = k_p
         self.k_i = k_i
         self.k_d = k_d
@@ -434,11 +327,12 @@ class LazyQuantileNorm(nn.Module):
 
     def __init__(self, quantiles=100, momentum=.001,
                  track_running_stats=True, use_stats_for_train=True, boost=True, predefined=None, scale=True):
-        super(LazyQuantileNorm, self).__init__()
+        super().__init__()
 
         quantiles = torch.arange(quantiles) / (quantiles - 1)
 
-        self.n_quantiles = len(quantiles) if predefined is None else predefined.shape[-1]
+        n_quantiles = len(quantiles) if predefined is None else predefined.shape[-1]
+        self.register_buffer("n_quantiles", torch.FloatTensor([n_quantiles]))
 
         boundaries = None if predefined is None else predefined
         self.predefined = False if predefined is None else True
@@ -457,8 +351,13 @@ class LazyQuantileNorm(nn.Module):
 
     def forward(self, x):
 
+        dtype = x.dtype
+
         if not self.track_running_stats or self.boundaries is None:
-            boundaries = torch.quantile(x, self.quantiles, dim=0).transpose(0, 1)
+
+            boundaries = torch.quantile(x.float(), self.quantiles.float(), dim=0).transpose(0, 1)
+            boundaries = boundaries.type(dtype)
+
             if self.boundaries is None:
                 self.boundaries = boundaries
         else:
@@ -467,8 +366,8 @@ class LazyQuantileNorm(nn.Module):
                 q = self.quantiles.view(1, 1, -1)
                 b = self.boundaries.unsqueeze(0)
                 xv = x.unsqueeze(-1).detach()
-                q_th = (q * (xv - b) > (1 - q) * (b - xv)).float()
-                q_grad = (- q * q_th + (1 - q) * (1 - q_th)) * (~torch.isinf(xv)).float()
+                q_th = (q * (xv - b) > (1 - q) * (b - xv)).type(dtype)
+                q_grad = (- q * q_th + (1 - q) * (1 - q_th)) * (~torch.isinf(xv)).type(dtype)
                 q_grad = q_grad.sum(dim=0)
 
                 if self.boost:
@@ -480,6 +379,7 @@ class LazyQuantileNorm(nn.Module):
                 self.boundaries = self.boundaries - self.lr * torch.std(x, dim=0).unsqueeze(-1) * factor * q_grad
                 self.boundaries = self.boundaries.sort(dim=1).values
 
+
         if (self.training and self.use_stats_for_train) or (
                 not self.training and self.track_running_stats) or self.predefined:
             boundaries = self.boundaries
@@ -494,12 +394,12 @@ class LazyQuantileNorm(nn.Module):
 class BetterEmbedding(torch.nn.Module):
 
     def __init__(self, numerical_indices, categorical_indices, n_quantiles, n_categories, emb_dim,
-                 momentum=.001, track_running_stats=True, noise=0.2, n_tables=15, initial_mask=1.,
+                 momentum=.001, track_running_stats=True, n_tables=15, initial_mask=1.,
                  k_p=0.05, k_i=0.005, k_d=0.005, T=20, clip=0.005, quantile_resolution=1e-4,
                  use_stats_for_train=True, boost=True, flatten=False, quantile_embedding=True, tokenizer=True,
-                 qnorm_flag=False, kaiming_init=False, init_spline_equally=True):
+                 qnorm_flag=False, kaiming_init=False, init_spline_equally=True, sparse=True, spline=True):
 
-        super(BetterEmbedding, self).__init__()
+        super().__init__()
 
         self.categorical_indices = categorical_indices
         self.numerical_indices = numerical_indices
@@ -510,6 +410,7 @@ class BetterEmbedding(torch.nn.Module):
         n_categories = n_categories + 1
         cat_offset = n_categories.cumsum(0) - n_categories
         self.register_buffer("cat_offset", cat_offset.unsqueeze(0))
+        self.register_buffer("null_emb_cat", torch.FloatTensor(1, 0, emb_dim))
 
         self.flatten = flatten
         self.n_tables = n_tables
@@ -520,9 +421,9 @@ class BetterEmbedding(torch.nn.Module):
         self.br = initial_mask
 
         if len(categorical_indices):
-            self.emb_cat = nn.Embedding(1 + self.n_emb * n_tables, emb_dim, sparse=True)
+            self.emb_cat = nn.Embedding(1 + self.n_emb * n_tables, emb_dim, sparse=sparse)
         else:
-            self.emb_cat = lambda x: torch.FloatTensor(len(x), 0, emb_dim).to(x.device)
+            self.emb_cat = lambda x: self.null_emb_cat.repeat(len(x), 1, 1)
 
         if init_spline_equally:
 
@@ -533,8 +434,10 @@ class BetterEmbedding(torch.nn.Module):
         else:
             weights = None
 
-        self.emb_num = SplineEmbedding(n_feature_num, n_quantiles, emb_dim, n_tables=n_tables,
-                                       enable=quantile_embedding, init_weights=weights)
+        self.emb_num = None
+        if spline:
+            self.emb_num = SplineEmbedding(n_feature_num, n_quantiles, emb_dim, n_tables=n_tables,
+                                       enable=quantile_embedding, init_weights=weights, sparse=sparse)
 
         self.llr = LazyLinearRegression(n_quantiles, lr=0.001)
         self.lambda_llr = 0
@@ -543,10 +446,12 @@ class BetterEmbedding(torch.nn.Module):
         else:
             self.pid_llr = PID(k_p=1e-4, k_i=0, k_d=0, T=T, clip=0)
 
-        self.qnorm = LazyQuantileNorm(quantiles=int(1 / quantile_resolution), momentum=momentum,
+        self.qnorm = None
+        if qnorm_flag:
+            self.qnorm = LazyQuantileNorm(quantiles=int(1 / quantile_resolution), momentum=momentum,
                                       track_running_stats=track_running_stats,
                                       use_stats_for_train=use_stats_for_train, boost=boost)
-        self.qnorm_flag = qnorm_flag
+
         self.tokenizer = tokenizer
         if tokenizer:
             self.weight = nn.Parameter(torch.empty((1, n_features, emb_dim)))
@@ -563,20 +468,23 @@ class BetterEmbedding(torch.nn.Module):
         else:
             nn.init.normal_(self.bias)
 
+        self.emb_dim = emb_dim
+
     def step(self, train_loss, val_loss):
 
         self.br = min(max(0, self.br + self.pid((train_loss - val_loss) / val_loss)), 1)
         self.lambda_llr = min(max(0, self.lambda_llr - self.pid_llr((train_loss - val_loss) / val_loss)), 1)
-        print(f"br was changed to {self.br}")
-        print(f"lambda_llr was changed to {self.lambda_llr}")
+        logger.info(f"br was changed to {self.br}")
+        logger.info(f"lambda_llr was changed to {self.lambda_llr}")
 
     def get_llr(self):
         return self.llr(self.emb_num.emb.weight) * self.lambda_llr
 
     def forward(self, x, ensemble=True):
 
-        x_num = x[:, self.numerical_indices].float()
-        if self.qnorm_flag:
+        x_num = x[:, self.numerical_indices]
+
+        if self.qnorm is not None:
             x_num = self.qnorm(x_num)
         x_cat = x[:, self.categorical_indices].long()
 
@@ -591,29 +499,59 @@ class BetterEmbedding(torch.nn.Module):
             mask_cat = 1
 
         if self.training:
-            rand_table = torch.randint(self.n_tables, size=(1, 1)).to(x_cat.device)
+            rand_table = torch.randint(self.n_tables, size=(1, 1), device=x_cat.device)
         else:
-            rand_table = torch.randint(self.n_tables, size=(len(x), 1)).to(x_cat.device)
+            rand_table = torch.randint(self.n_tables, size=(len(x), 1), device=x_cat.device)
 
         x_cat = (x_cat + 1) * mask_cat + self.cat_offset + self.n_emb * rand_table
 
         e_cat = self.emb_cat(x_cat)
-        e_num = self.emb_num(x_num, mask_num, rand_table)
+
+        if self.emb_num is not None:
+            e_num = self.emb_num(x_num, mask_num, rand_table)
+        else:
+            e_num = torch.zeros(*x_num.shape, self.emb_dim, device=x_num.device, dtype=x_num.dtype)
 
         e = torch.cat([e_cat, e_num], dim=1)
         e = e + self.bias
-        #         print(e[0,100])
+
         if self.tokenizer:
             x = torch.cat([torch.zeros_like(x_cat), x_num * mask_num], dim=1)
             y = self.weight * x.unsqueeze(-1)
             e = e + y
-        #             print(e[0,100])
         if self.flatten:
             e = e.view(len(e), -1)
 
         return e
 
 
+class mySequential(nn.Sequential):
+    def forward(self, *input):
+        for module in self._modules.values():
+            input = module(*input)
+        return input
+
+
+class GBN(torch.nn.Module):
+    """
+    Ghost Batch Normalization
+    https://arxiv.org/abs/1705.08741
+    """
+
+    def __init__(self, input_dim, virtual_batch_size=128, momentum=0.01):
+        super().__init__()
+
+        self.input_dim = input_dim
+        self.virtual_batch_size = virtual_batch_size
+        self.bn = nn.BatchNorm1d(self.input_dim, momentum=momentum)
+
+    def forward(self, x):
+        chunks = x.chunk(int(np.ceil(x.shape[0] / self.virtual_batch_size)), 0)
+        res = [self.bn(x_) for x_ in chunks]
+
+        return torch.cat(res, dim=0)
+
+
 class Sparsemax(nn.Module):
     """Sparsemax function."""
 
@@ -623,7 +561,7 @@ class Sparsemax(nn.Module):
         Args:
             dim (int, optional): The dimension over which to apply the sparsemax function.
         """
-        super(Sparsemax, self).__init__()
+        super().__init__()
 
         self.dim = -1 if dim is None else dim
 
@@ -690,6 +628,72 @@ class Sparsemax(nn.Module):
         return self.grad_input
 
 
+class BeamEnsemble(torch.nn.Module):
+
+    def __init__(self, net, n_ensembles, optimizer=None):
+        super().__init__()
+
+        if isinstance(net, nn.Module):
+            ensembles = []
+            for _ in range(n_ensembles):
+                new_net = copy_network(net)
+                reset_network(new_net)
+                ensembles.append(new_net)
+        else:
+            ensembles = [net() for _ in range(n_ensembles)]
+
+        self.ensembles = nn.ModuleList(ensembles)
+        self.n_ensembles = n_ensembles
+
+        self.optimizers = None
+        if optimizer is not None:
+            self.optimizers = self.set_optimizers(optimizer)
+
+        self.optimizer = None
+        self.net = None
+        self.active_model = None
+
+    def set_optimizers(self, optimizer):
+
+        try:
+            self.optimizers = [optimizer(net) for net in self.ensembles]
+        except TypeError:
+            self.optimizers = [optimizer(net.paramters()) for net in self.ensembles]
+
+        return self.optimizers
+
+    def __len__(self):
+        return self.n_ensembles
+
+    def forward(self, x, reduction='mean', index=None):
+
+        if self.training:
+            if index is None:
+                    index = random.randint(0, self.n_ensembles-1)
+
+            self.net = self.ensembles[index]
+            if self.optimizers is not None:
+                self.optimizer = self.optimizers[index]
+            self.active_model = index
+
+            y = self.net(x)
+
+        else:
+            self.optimizer = None
+            self.net = None
+            self.active_model = None
+
+            y = [net(x) for net in self.ensembles]
+            if reduction == 'mean':
+                y = torch.stack(y).mean(dim=0)
+            elif reduction == 'none':
+                y = torch.stack(y)
+            else:
+                raise NotImplementedError
+
+        return y
+
+
 class FeatureHasher(object):
 
     def __init__(self, num_embeddings, embedding_dim, n_classes=None, distribution='uniform', seed=None, device='cpu'):
@@ -710,16 +714,170 @@ class FeatureHasher(object):
         return self.weight[x]
 
 
+def beam_weights_initializer(net, black_list=None, white_list=None, zero_bias=True,
+                             method=None, gain=None, nonlinearity='relu', method_argv=None,
+                             nonlinearity_argv=None, method_linear=None, method_linear_argv=None,
+                             method_conv=None, method_conv_argv=None, method_embedding=None, method_embedding_argv=None,
+                             temperature=1):
+    """
+
+    @param net:
+    @param black_list:
+    @param white_list:
+    @param zero_bias:
+    @param method: [kaiming_uniform, kaiming_normal, xavier_uniform, xavier_normal, orthogonal, trunc_normal, sparse]
+    @param gain:
+    @param nonlinearity:
+    @param nonlinearity_argv:
+    @return:
+    """
+
+    def bias_init(m):
+        if zero_bias:
+            nn.init.zeros_(m.bias)
+        else:
+            fan_in, _ = nn.init._calculate_fan_in_and_fan_out(m.weight)
+            if fan_in != 0:
+                bound = 1 / math.sqrt(fan_in)
+                nn.init.uniform_(m.bias, -bound, bound)
+
+    if gain is None:
+        if nonlinearity_argv is None:
+            nonlinearity_argv = {}
+        gain = nn.init.calculate_gain(nonlinearity, **nonlinearity_argv)
+
+    # bert initializes embedding of 768 with .02 std and pytorch initializes with 1. stdout
+    # so by default we use something in the middle which is also more reasonable for embedding of ~256 dimensions.
+    if method_embedding is None:
+        method_embedding = partial(nn.init.normal_, mean=0., std=.2)
+    else:
+        if method_embedding_argv is None:
+            method_embedding_argv = {}
+        method_embedding = partial(getattr(nn.init, f"{method_embedding}_"), **method_embedding_argv)
+
+    if method is not None:
+        if method_argv is None:
+            method_argv = {}
+        method = partial(getattr(nn.init, f"{method}_"), **method_argv)
+
+    if method_conv is not None:
+        if method_conv_argv is None:
+            method_conv_argv = {}
+        method_conv = partial(getattr(nn.init, f"{method_conv}_"), **method_conv_argv)
+    elif method is not None:
+        method_conv = method
+    else:
+        method_conv = partial(nn.init.xavier_uniform_, gain=gain)
+
+    if method_linear is not None:
+        if method_linear_argv is None:
+            method_linear_argv = {}
+        method_linear = partial(getattr(nn.init, f"{method_conv}_"), **method_linear_argv)
+    elif method is not None:
+        method_linear = method
+    else:
+        method_linear = partial(nn.init.xavier_uniform_, gain=gain)
+
+    if black_list is None:
+        black_list = []
+    if white_list is None:
+        # white_list = [n for n, p in net.named_parameters()]
+        white_list = ['']
+
+    def valid_param(n, suffix=None):
+        if suffix is not None:
+            n = f"{n}.{suffix}"
+        return any([nw in n for nw in white_list]) and not any([nb in n for nb in black_list])
+
+    for n, m in net.named_modules():
+        if len(list(m.children())) or not valid_param(n):
+            continue
+
+        if 'Norm' in str(type(m)):
+            if valid_param(n, suffix='weight') and hasattr(m, 'weight'):
+                nn.init.ones_(m.weight)
+            if valid_param(n, suffix='bias') and hasattr(m, 'bias'):
+                nn.init.ones_(m.bias)
+        elif 'Embedding' in str(type(m)):
+            if valid_param(n, suffix='weight') and hasattr(m, 'weight'):
+                method_embedding(m.weight)
+
+        elif 'Linear' in str(type(m)):
+            if valid_param(n, suffix='weight') and hasattr(m, 'weight'):
+                fan_in, fan_out = nn.init._calculate_fan_in_and_fan_out(m.weight)
+                if fan_in > 1 and fan_out > 1:
+                    method_linear(m.weight)
+                else:
+                    with torch.no_grad():
+                        m.weight.normal_()
+                        m.weight.data = torch.exp(m.weight.data / (temperature * math.sqrt(torch.numel(m.weight))))
+                        m.weight.data = m.weight.data / m.weight.data.sum()
+
+            if valid_param(n, suffix='bias') and hasattr(m, 'bias'):
+                bias_init(m)
+        elif 'Conv' in str(type(m)):
+            if valid_param(n, suffix='weight') and hasattr(m, 'weight'):
+                fan_in, fan_out = nn.init._calculate_fan_in_and_fan_out(m.weight)
+                if fan_in > 1 and fan_out > 1:
+                    method_conv(m.weight)
+                else:
+                    with torch.no_grad():
+                        m.weight.normal_()
+                        m.weight.data = torch.exp(m.weight.data / (temperature * math.sqrt(torch.numel(m.weight))))
+                        m.weight.data = m.weight.data / m.weight.data.sum()
+            if valid_param(n, suffix='bias') and hasattr(m, 'bias'):
+                bias_init(m)
+        else:
+            if len(list(m.parameters())) > 0:
+                logger.warning(f"Beam weight initializer does not support layer type: {n}")
+
+
+def reset_network(net):
+    prev_hash = {n: hash_tensor(p) for n, p in net.named_parameters()}
+    for n, m in net.named_modules():
+        if hasattr(m, 'reset_parameters'):
+            m.reset_parameters()
+    for n, p in net.named_parameters():
+        if prev_hash[n] == hash_tensor(p):
+            logger.warning(f"Parameter {n} was not reset. Check if its nn.Module supports .reset_parameters()")
+
+
+def free_network_params(*nets):
+    for net in nets:
+        for p in net.parameters():
+            p.requires_grad = True
+
+
+def freeze_network_params(*nets):
+    for net in nets:
+        for p in net.parameters():
+            p.requires_grad = False
+
+
+def copy_network(net):
+    return copy.deepcopy(net)
+
+
+def soft_target_update(source, target, tau):
+    for target_param, param in zip(target.parameters(), source.parameters()):
+        target_param.data.copy_(target_param.data * tau + param.data * (1.0 - tau))
+
+
+def target_copy(source, target):
+    for target_param, param in zip(target.parameters(), source.parameters()):
+        target_param.data.copy_(param.data)
+
+
 def reset_networks_and_optimizers(networks=None, optimizers=None):
     if networks is not None:
-        net_iter = networks.keys() if issubclass(type(networks), dict) else range(len(networks))
+        net_iter = networks.keys() if isinstance(networks, dict) else range(len(networks))
         for i in net_iter:
             for n, m in networks[i].named_modules():
                 if hasattr(m, 'reset_parameters'):
                     m.reset_parameters()
 
     if optimizers is not None:
-        opt_iter = optimizers.keys() if issubclass(type(optimizers), dict) else range(len(optimizers))
+        opt_iter = optimizers.keys() if isinstance(optimizers, dict) else range(len(optimizers))
         for i in opt_iter:
             opt = optimizers[i]
 
